import os

import click

from . import __version__
from .log import configure_logger
from .server import run_forever
from .settings import load_settings


@click.group()
@click.version_option(version=__version__)
def cli():
    pass


@cli.command()
@click.option('--env-path',
              type=click.Path(exists=True, readable=True, resolve_path=True),
              help='Dot env file or directory path.')
@click.option('--verbose', is_flag=True, help='Enable more traces.')
@click.option('--logger-config',
              type=click.Path(exists=False, readable=True, resolve_path=True),
              help='Python logger JSON definition.')
def run(env_path: str = None, verbose: bool = False,
        logger_config: str = None):
    """
    Runs the application.
    """
    if verbose:
        os.environ["CHAOSPLATFORM_DEBUG"] = "1"
    config = load_settings(env_path)

    configure_logger(logger_config, config)
    run_forever(config)



import logging
from logging import StreamHandler
from typing import Any, Dict, Tuple

from chaosplt_grpc import create_grpc_server, start_grpc_server, \
    stop_grpc_server
import cherrypy
from flask import Flask
from flask_caching import Cache
from grpc import Server

from .cache import setup_cache
from .service import initialize_services, shutdown_services, Services
from .storage import SchedulingStorage, initialize_storage, shutdown_storage
from .views.api import create_api, cleanup_api, serve_api
from .views.web import create_app, cleanup_app, serve_app

__all__ = ["initialize_all", "release_all", "run_forever"]
logger = logging.getLogger("chaosplatform")


def initialize_all(config: Dict[str, Any], web_app: Flask = None,
                   api_app: Flask = None, services: Services = None,
                   grpc_server: Server = None, web_cache: Cache = None,
                   api_cache: Cache = None,
                   web_mount_point: str = "/scheduling",
                   api_mount_point: str = "/api/v1/scheduling",
                   access_log_handler: StreamHandler = None) \
                   -> Tuple[Flask, Flask, Services, Server, SchedulingStorage]:
    access_log_handler = access_log_handler or logging.StreamHandler()
    logger.info("Initializing scheduling service resources")

    embedded = True
    if not services:
        embedded = False
        services = Services()

    storage = initialize_storage(config)
    if embedded:
        services.scheduling = storage

    initialize_services(services, config)

    if not web_app:
        web_app = create_app(config)
        web_cache = setup_cache(web_app)
    serve_app(
        web_app, web_cache, services, storage, config, web_mount_point,
        log_handler=access_log_handler)

    if not api_app:
        api_app = create_api(config)
        api_cache = setup_cache(api_app)
    serve_api(
        api_app, api_cache, services, storage, config, api_mount_point,
        log_handler=access_log_handler)

    if not grpc_server:
        srv_addr = config["GRPC_LISTEN_ADDR"]
        if srv_addr:
            grpc_server = create_grpc_server(srv_addr)
            start_grpc_server(grpc_server)
            logger.info("gRPC server started on {}".format(srv_addr))

    return (web_app, api_app, services, grpc_server, storage)


def release_all(web_app: Flask, api_app: Flask, services: Services,
                grpc_server: Server, storage: SchedulingStorage):
    logger.info("Releasing scheduling service resources")
    if grpc_server:
        logger.info("gRPC server stopping")
        stop_grpc_server(grpc_server)
        logger.info("gRPC server stopped")
    cleanup_app(web_app)
    cleanup_api(api_app)
    shutdown_services(services)
    shutdown_storage(storage)


def run_forever(config: Dict[str, Any]):
    """
    Run and block until a signal is sent to the process.

    The application, services or gRPC server are all created and initialized
    when the application starts.
    """
    def run_stuff(config: Dict[str, Any]):
        resources = initialize_all(config)
        cherrypy.engine.subscribe(
            'stop', lambda: release_all(*resources),
            priority=20)

    cherrypy.engine.subscribe(
        'start', lambda: run_stuff(config), priority=80)

    cherrypy.engine.signals.subscribe()
    cherrypy.engine.start()
    cherrypy.engine.block()



# -*- coding: utf-8 -*-
import glob
import logging
import os
from typing import Any, Dict

import cherrypy
from dotenv import load_dotenv

__all__ = ["load_settings"]
logger = logging.getLogger("chaosplatform")


def load_settings(env_path: str) -> Dict[str, Any]:
    """
    Load settings from the environment:

    * if `env_path` is a file, read it
    * if `env_path` is a directory, load, all its `*.env` files
    """
    if os.path.isdir(env_path):
        pattern = os.path.join(env_path, '**', '.env')
        for env_file in glob.iglob(pattern, recursive=True):
            logger.info("Loading: {}".format(env_file))
            load_dotenv(dotenv_path=env_file)
    else:
        logger.info("Loading: {}".format(env_path))
        load_dotenv(dotenv_path=env_path)

    debug = True if os.getenv('CHAOSPLATFORM_DEBUG') else False
    cherrypy.config.update({
        'server.socket_host': os.getenv('SERVER_LISTEN_ADDR'),
        'server.socket_port': int(os.getenv('SERVER_LISTEN_PORT', 8080)),
        'engine.autoreload.on': False,
        'log.screen': debug,
        'log.access_file': '',
        'log.error_file': '',
        'environment': '' if debug else 'production',
        'tools.proxy.on': True,
        'tools.proxy.base': os.getenv('HTTP_PROXY_BASE')
    })

    config = {
        'GRPC_LISTEN_ADDR': os.getenv("GRPC_LISTEN_ADDR"),
        'db': {
            'uri': os.getenv("DATABASE_URI"),
            'debug': debug,
        },
        'debug': debug,
        'grpc': {
            'scheduler': {
                'addr': os.getenv('SCHEDULER_SERVICE_GRPC_ADDR')
            }
        },
        'jwt': {
            'secret_key': os.getenv("JWT_SECRET_KEY"),
            'public_key': os.getenv("JWT_PUBLIC_KEY"),
            'algorithm': os.getenv("JWT_ALGORITHM", "HS256"),
            'identity_claim_key': os.getenv("JWT_IDENTITY_CLAIM", "identity"),
            'user_claims_key': os.getenv("JWT_USER_CLAIMS", "user_claims"),
            'access_token_expires': os.getenv("JWT_ACCESS_TOKEN_EXPIRES"),
            'refresh_token_expires': os.getenv(
                "JWT_REFRESH_TOKEN_EXPIRES", 30*24*60*60),
            'user_claims_in_refresh_token': os.getenv(
                "JWT_CLAIMS_IN_REFRESH_TOKEN")
        }
    }

    return config



from collections import OrderedDict
from typing import Dict
from uuid import UUID

from chaosplt_scheduling.schemas import upload_scheduling_schema


def test_scheduling_request(scheduling: Dict[str, str]):
    s = upload_scheduling_schema.load(scheduling)
    assert str(s["org"]) == scheduling["org"]
    assert str(s["workspace"]) == scheduling["workspace"]
    assert str(s["token"]) == scheduling["token"]
    assert str(s["experiment"]) == scheduling["experiment"]
    assert "repeat" not in s
    assert "interval" not in s
    assert "cron" not in s
    assert "settings" not in s
    assert "configuration" not in s
    assert "secrets" not in s


def test_cron_scheduling_request(scheduling: Dict[str, str]):
    scheduling["cron"] = "* 8 * * *"
    s = upload_scheduling_schema.load(scheduling)
    assert str(s["org"]) == scheduling["org"]
    assert str(s["workspace"]) == scheduling["workspace"]
    assert str(s["token"]) == scheduling["token"]
    assert str(s["experiment"]) == scheduling["experiment"]
    assert s["cron"] == "* 8 * * *"
    assert "repeat" not in s
    assert "interval" not in s
    assert "settings" not in s
    assert "configuration" not in s
    assert "secrets" not in s


def test_cron_scheduling_with_repeat_request(scheduling: Dict[str, str]):
    scheduling["cron"] = "* 8 * * *"
    scheduling["repeat"] = 5
    s = upload_scheduling_schema.load(scheduling)
    assert str(s["org"]) == scheduling["org"]
    assert str(s["workspace"]) == scheduling["workspace"]
    assert str(s["token"]) == scheduling["token"]
    assert str(s["experiment"]) == scheduling["experiment"]
    assert s["cron"] == "* 8 * * *"
    assert s["repeat"] == 5
    assert "interval" not in s
    assert "settings" not in s
    assert "configuration" not in s
    assert "secrets" not in s


def test_interval_scheduling_request(scheduling: Dict[str, str]):
    scheduling["interval"] = 10
    s = upload_scheduling_schema.load(scheduling)
    assert str(s["org"]) == scheduling["org"]
    assert str(s["workspace"]) == scheduling["workspace"]
    assert str(s["token"]) == scheduling["token"]
    assert str(s["experiment"]) == scheduling["experiment"]
    assert s["interval"] == 10
    assert "repeat" not in s
    assert "cron" not in s
    assert "settings" not in s
    assert "configuration" not in s
    assert "secrets" not in s


def test_interval_with_repeat_scheduling_request(scheduling: Dict[str, str]):
    scheduling["interval"] = 10
    scheduling["repeat"] = 4
    s = upload_scheduling_schema.load(scheduling)
    assert str(s["org"]) == scheduling["org"]
    assert str(s["workspace"]) == scheduling["workspace"]
    assert str(s["token"]) == scheduling["token"]
    assert str(s["experiment"]) == scheduling["experiment"]
    assert s["interval"] == 10
    assert s["repeat"] == 4
    assert "cron" not in s
    assert "settings" not in s
    assert "configuration" not in s
    assert "secrets" not in s



import os
import sys

from pkg_resources import Distribution, EntryPoint, iter_entry_points, \
    working_set

from chaosplt_scheduling.storage import initialize_storage, shutdown_storage


def test_loading_external_storage_implementation():
    try:
        distribution = Distribution(os.path.join(
                os.path.dirname(__file__), "fixtures"))
        entry_point = EntryPoint.parse(
            'scheduling = fake_storage:MySchedulingStorage', dist=distribution)
        distribution._ep_map = {
            'chaosplatform.storage': {'scheduling': entry_point}}
        working_set.add(distribution)

        storage = initialize_storage(config={"db":{"uri": "sqlite:///"}})
        assert storage.__class__.__name__ == "MySchedulingStorage"
        assert storage.some_flag == True

        shutdown_storage(storage)
        assert storage.some_flag == False

    finally:
        distribution._ep_map.clear()



from datetime import datetime
import os
import random
import string
from typing import Any, Dict
from unittest.mock import MagicMock
import uuid
from uuid import UUID

import cherrypy
from flask import  Flask, request
from flask_caching import Cache
from flask_login import login_required, login_user, logout_user
import logging
from logging import StreamHandler
import pytest
from sqlalchemy.sql import text as sa_text

from chaosplt_relational_storage.db import RelationalStorage
from chaosplt_scheduling.cache import setup_cache
from chaosplt_scheduling.log import http_requests_logger
from chaosplt_scheduling.views.api import create_api, cleanup_api, serve_api, \
    register_api
from chaosplt_scheduling.views.web import create_app, cleanup_app, serve_app, \
    register_views
from chaosplt_scheduling.service import Services, initialize_services
from chaosplt_scheduling.settings import load_settings
from chaosplt_scheduling.storage import initialize_storage, shutdown_storage, \
    SchedulingStorage

cur_dir = os.path.abspath(os.path.dirname(__file__))
fixtures_dir = os.path.join(cur_dir, "fixtures")
env_path = os.path.join(fixtures_dir, '.env')


@pytest.fixture
def config():
    return load_settings(env_path)


@pytest.fixture(autouse=True)
def storage(config: Dict[str, Any]) -> SchedulingStorage:
    s = initialize_storage(config)
    yield s
    # clear out the table from all data
    # dropping seems to cause issues
    s.driver.engine.execute(
        sa_text("DELETE FROM schedule;").execution_options(
            autocommit=True))


@pytest.fixture
def stream_handler() -> StreamHandler:
    return StreamHandler()


@pytest.fixture
def scheduler_service():
    return MagicMock()


@pytest.fixture
def services(config: Dict[str, Any], scheduler_service) -> Services:
    svc = Services()
    svc.scheduler = scheduler_service
    initialize_services(svc, config)
    return svc


@pytest.fixture
def app(config: Dict[str, Any], services: Services, storage: SchedulingStorage,
        stream_handler: StreamHandler) -> Flask:
    application = create_app(config)
    cache = setup_cache(application)
    serve_app(
        application, cache, services, storage, config, "/scheduling",
        log_handler=stream_handler)
    return application


@pytest.fixture
def api(app: Flask, config: Dict[str, Any], services: Services,
        storage: SchedulingStorage, stream_handler: StreamHandler) -> Flask:
    application = create_api(config)
    cache = setup_cache(application)
    wsgiapp = http_requests_logger(application, stream_handler)
    cherrypy.tree.graft(wsgiapp, "/api/v1")
    serve_api(
        application, cache, services, storage, config, "/api/v1/scheduling",
        log_handler=stream_handler)
    return application


@pytest.fixture
def app_cache(app: Flask) -> Cache:
    return Cache(app, config=app.config)


@pytest.fixture
def api_cache(api: Flask) -> Cache:
    return Cache(api, config=api.config)


@pytest.fixture
def user_id() -> UUID:
    return uuid.uuid4() #UUID("1d77476b-f11d-433b-b7ee-bcbf25862051")


@pytest.fixture
def org_id() -> UUID:
    return uuid.uuid4()


@pytest.fixture
def workspace_id() -> UUID:
    return uuid.uuid4()


@pytest.fixture
def experiment_id() -> UUID:
    return uuid.uuid4()


@pytest.fixture
def token_id() -> UUID:
    return uuid.uuid4()


@pytest.fixture
def scheduled() -> str:
    return "{}Z".format(datetime.utcnow().isoformat())


@pytest.fixture
def app_client(app: Flask, services: Services, username: str, name: str,
               email: str):
    app.config['TESTING'] = True
    app.testing = True

    with app.test_request_context():
        user = services.account.registration.create(
            username, name, email)
        login_user(user, remember=True)
        services.account.registration.get(user_id)

    with app.test_client() as client:
        yield client


@pytest.fixture
def api_client(api: Flask, services: Services, username: str, name: str,
               email: str):
    api.config['TESTING'] = True
    api.testing = True

    with api.test_request_context():
        user = services.account.registration.create(
            username, name, email)
        login_user(user, remember=True)
        services.account.registration.get(user_id)

    with api.test_client() as client:
        yield client


@pytest.fixture
def scheduling(org_id: UUID, workspace_id: UUID, experiment_id: UUID,
               token_id: UUID, scheduled: str) -> Dict[str, str]:
    return {
        "org": str(org_id),
        "workspace": str(workspace_id),
        "experiment": str(experiment_id),
        "token": str(token_id),
        "scheduled": scheduled
    }



from flask import Flask
from flask_marshmallow import Marshmallow
from marshmallow import fields

__all__ = ["ma", "schedule_schema", "upload_scheduling_schema",
           "setup_schemas"]

ma = Marshmallow()


def setup_schemas(app: Flask):
    return ma.init_app(app)


class ScheduleSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=True)
    user_id = fields.UUID(required=True)
    org_id = fields.UUID(required=True)
    workspace_id = fields.UUID(required=True)
    experiment_id = fields.UUID(required=True)
    token_id = fields.UUID(required=True)
    job_id = fields.UUID(allow_none=True)
    scheduled = fields.DateTime()
    repeat = fields.Integer(allow_none=True)
    interval = fields.Integer(allow_none=True)
    cron = fields.String(allow_none=True)
    results = fields.List(fields.Dict(keys=fields.String()), allow_none=True)
    url = ma.AbsoluteURLFor('scheduling.get', schedule_id='<id>')
    links = ma.Hyperlinks({
        'self': ma.URLFor('scheduling.get', schedule_id='<id>')
    })


class UploadSchedulingSchema(ma.Schema):
    org = fields.UUID(required=True)
    workspace = fields.UUID(required=True)
    experiment = fields.UUID(required=True)
    token = fields.UUID(required=True)
    scheduled = fields.String(required=True)
    repeat = fields.Integer(allow_none=True)
    interval = fields.Integer(allow_none=True)
    cron = fields.String(allow_none=True)
    settings = fields.Dict(keys=fields.String(), allow_none=True)
    configuration = fields.Dict(keys=fields.String(), allow_none=True)
    secrets = fields.Dict(keys=fields.String(), allow_none=True)


schedule_schema = ScheduleSchema()
schedules_schema = ScheduleSchema(many=True)
upload_scheduling_schema = UploadSchedulingSchema()



from typing import Any, Dict

import attr

from .scheduler import SchedulerService


__all__ = ["initialize_services", "shutdown_services", "Services"]


@attr.s
class Services:
    scheduler: object = attr.ib(default=None)


def initialize_services(services: Services, config: Dict[str, Any]):
    """
    Create and initialize services access:

    * scheduler
    """
    if not services.scheduler:
        services.scheduler = SchedulerService(config)


def shutdown_services(services: Services):
    """
    Release resources used by the services.

    They should not be assumed to be functional after this call.
    """
    if services.scheduler:
        services.scheduler.release()



import pytest

from chaosplt_scheduling.storage.interface import BaseSchedulingService

def test_cannot_instanciate_scheduling_interface():
    try:
        BaseSchedulingService()
    except TypeError as e:
        return
    else:
        pytest.fail("BaseSchedulingService should remain abstract")



import logging
import logging.config
from logging import StreamHandler
import pkgutil
from typing import Any, Dict

from flask import Flask
from requestlogger import ApacheFormatter, WSGILogger
import simplejson as json

__all__ = ["configure_logger", "http_requests_logger"]

logger = logging.getLogger("chaosplatform")


def configure_logger(log_conf_path: str, config: Dict[str, Any]):
    if log_conf_path:
        with open(log_conf_path) as f:
            log_conf = json.load(f)
    else:
        log_conf = json.loads(
            pkgutil.get_data('chaosplt_auth', 'logging.json'))
    logging.config.dictConfig(log_conf)
    if config["debug"]:
        [h.setLevel(logging.DEBUG) for h in logger.handlers]
        logger.setLevel(logging.DEBUG)
    logger.debug("Logger configured")


def http_requests_logger(app: Flask,
                         stream_handler: StreamHandler = None) -> WSGILogger:
    if not stream_handler:
        stream_handler = StreamHandler()
    return WSGILogger(
        app.wsgi_app, [stream_handler], ApacheFormatter(), propagate=False)



# -*- coding: utf-8 -*-
import logging
from logging import StreamHandler
import os
from typing import Any, Dict

from flask import Blueprint, Flask, after_this_request, request, Response
from flask_caching import Cache

from chaosplt_scheduling.auth import setup_jwt, setup_login
from chaosplt_scheduling.schemas import setup_schemas
from chaosplt_scheduling.service import Services
from chaosplt_scheduling.storage import SchedulingStorage

from .scheduling import api as scheduling_api

__all__ = ["create_api", "cleanup_api", "register_api"]


def create_api(config: Dict[str, Any]) -> Flask:
    """
    Create the API application for the scheduling API endpoint.
    """
    app = Flask(__name__)

    app.url_map.strict_slashes = False
    app.debug = True if os.getenv('CHAOSPLATFORM_DEBUG') else False

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
    app.config["SQLALCHEMY_DATABASE_URI"] = os.getenv("DATABASE_URI")
    app.config["JWT_SECRET_KEY"] = os.getenv("JWT_SECRET_KEY")
    app.secret_key = app.config["SECRET_KEY"]

    app.config["GRPC_ADDR"] = "{}:{}".format(
        os.getenv("GRPC_LISTEN_ADDR"), os.getenv("GRPC_LISTEN_PORT"))

    app.config["CACHE_TYPE"] = os.getenv("CACHE_TYPE", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = os.getenv("CACHE_REDIS_HOST")
        app.config["CACHE_REDIS_PORT"] = os.getenv("CACHE_REDIS_PORT", 6379)

    setup_jwt(app)
    setup_schemas(app)
    setup_login(app, from_jwt=True)

    return app


def cleanup_api(app: Flask):
    pass


def serve_api(app: Flask, cache: Cache, services: Services,
              storage: SchedulingStorage, config: Dict[str, Any],
              mount_point: str = '/api/v1/scheduling',
              log_handler: StreamHandler = None):
    """
    Serve the scheduling API application over HTTP.
    """
    register_api(app, cache, services, storage, mount_point)


###############################################################################
# Internals
###############################################################################
def register_api(app: Flask, cache: Cache, services: Services,
                 storage: SchedulingStorage, mount_point: str):
    patch_request(scheduling_api, services, storage)
    app.register_blueprint(scheduling_api, url_prefix=mount_point)


def patch_request(bp: Blueprint, services, storage: SchedulingStorage):
    @bp.before_request
    def prepare_request():
        request.services = services
        request.storage = storage

        @after_this_request
        def clean_request(response: Response):
            request.services = None
            request.storage = None
            return response



import os.path

import cherrypy
from flask import Flask

from chaosplt_scheduling.settings import load_settings

cur_dir = os.path.abspath(os.path.dirname(__file__))
fixtures_dir = os.path.join(cur_dir, "fixtures")
env_path = os.path.join(fixtures_dir, '.env')


def test_load_single_dot_env():
    config = load_settings(env_path)
    assert config.get('debug') == True


def test_load_from_directory():
    config = load_settings(fixtures_dir)
    assert config.get('debug') == True
    assert cherrypy.server.socket_port == 11091


def test_configuring_app(app: Flask):
    assert app.config["SECRET_KEY"] == "whatever"
    assert app.secret_key == "whatever"
    assert app.config["JWT_SECRET_KEY"] == "jwt_whatever"
    assert app.config["GRPC_LISTEN_ADDR"] == "0.0.0.0:50053"
    assert app.config["CACHE_TYPE"] == "simple"



from typing import Any, Dict, NoReturn

from chaosplt_relational_storage import get_storage, \
    configure_storage, release_storage
import pkg_resources

from .concrete import SchedulingService
from .interface import BaseSchedulingStorage

__all__ = ["initialize_storage", "shutdown_storage", "SchedulingStorage"]


class SchedulingStorage(BaseSchedulingStorage, SchedulingService):
    def __init__(self, config: Dict[str, Any]):
        BaseSchedulingStorage.__init__(self)

        driver = get_storage(config)
        configure_storage(driver)
        SchedulingService.__init__(self, driver)

    def release(self) -> NoReturn:
        """
        Release the resources used by the scheduling storage

        Do not use the storage after this method has been called.
        """
        release_storage(self.driver)


def initialize_storage(config: Dict[str, Any]) -> SchedulingStorage:
    """
    Initialize the storage.

    Look first for any plugin implementing the `'chaosplatform.storage'`
    entrypoint group and load it if any is found for the `"scheduling"` name.

    Otherwise, create an instance of `SchedulingStorage` from this module.
    """
    for plugin in pkg_resources.iter_entry_points('chaosplatform.storage'):
        if plugin.name == "scheduling":
            service_class = plugin.load()
            return service_class(config)

    return SchedulingStorage(config)


def shutdown_storage(storage: SchedulingStorage) -> NoReturn:
    """
    Release all resources used by the scheduling storage
    """
    storage.release()



# -*- coding: utf-8 -*-
from uuid import UUID

from dateparser import parse
from flask import abort, Blueprint, jsonify, request
from flask_login import current_user, login_required
from marshmallow import ValidationError
import simplejson as json

from chaosplt_scheduling.schemas import schedule_schema, \
        upload_scheduling_schema

__all__ = ["api"]

api = Blueprint("scheduling", __name__)


@api.route('', methods=['POST'])
@login_required
def new():
    user_id = current_user.id
    try:
        payload = upload_scheduling_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422

    org_id = payload["org"]
    workspace_id = payload["workspace"]
    experiment_id = payload["experiment"]
    token_id = payload["token"]
    scheduled_at = payload["scheduled"]

    # periodicity
    interval = payload.get("interval", None)
    cron = payload.get("cron", None)
    repeat = payload.get("repeat", None)

    # context for the experiment
    settings = payload.get("settings", None)
    configuration = payload.get("configuration", None)
    secrets = payload.get("secrets", None)

    if interval and cron:
        return jsonify({
            "message": "Cannot use cron and interval at the same time"
        }), 422

    schedule_dt = parse(scheduled_at, settings={'TO_TIMEZONE': 'UTC'})
    if not schedule_dt:
        return jsonify({
            "message": "Invalid schedule date"
        }), 422

    token = request.services.auth.access_token.get(user_id, token_id)
    if not token:
        return jsonify({
            "message": "Invalid token"
        }), 422

    experiment = request.services.experiment.get_by_user(
        user_id, experiment_id)
    if not experiment:
        return jsonify({
            "message": "Invalid experiment"
        }), 422

    # store the schedule
    schedule = request.services.scheduling.create(
        user_id, org_id, workspace_id, experiment_id, token_id,
        scheduled_at, repeat, interval, cron, settings, configuration, secrets)

    # trigger the actual scheduler
    job_id = request.services.scheduler.schedule(
        str(schedule.id), str(user_id), str(org_id), str(workspace_id),
        str(schedule.experiment_id), str(token_id), token.access_token,
        scheduled_at, json.dumps(experiment.payload),
        interval, repeat, cron, settings, configuration, secrets)

    # update our records
    request.services.scheduling.job_started(schedule.id, job_id)
    schedule = request.services.scheduling.get(user_id, schedule.id)
    schedule.job_id = job_id

    return schedule_schema.jsonify(schedule)


@api.route('<uuid:schedule_id>', methods=['GET'])
@login_required
def get(schedule_id: UUID):
    user_id = current_user.id
    schedule = request.services.scheduling.get(user_id, schedule_id)
    if not schedule:
        return abort(404)
    return schedule_schema.jsonify(schedule)


@api.route('<uuid:schedule_id>', methods=['DELETE'])
@login_required
def delete(schedule_id: UUID):
    user_id = current_user.id
    schedule = request.services.scheduling.get(user_id, schedule_id)
    if schedule:
        request.services.scheduler.cancel(str(schedule.job_id))
        request.services.scheduling.delete(schedule_id)
    return jsonify(""), 204



# -*- coding: utf-8 -*-
import logging
from logging import StreamHandler
import os
from typing import Any, Dict

from flask import Blueprint, Flask, request, after_this_request, Response
from flask_caching import Cache

from chaosplt_scheduling.service import Services
from chaosplt_scheduling.storage import SchedulingStorage

__all__ = ["create_app", "cleanup_app", "serve_app"]


def create_app(config: Dict[str, Any]) -> Flask:
    app = Flask(__name__)

    app.url_map.strict_slashes = False
    app.debug = True if os.getenv('CHAOSPLATFORM_DEBUG') else False

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
    app.config["JWT_SECRET_KEY"] = os.getenv("JWT_SECRET_KEY")
    app.secret_key = app.config["SECRET_KEY"]

    app.config["GRPC_LISTEN_ADDR"] = os.getenv("GRPC_LISTEN_ADDR")

    app.config["CACHE_TYPE"] = os.getenv("CACHE_TYPE", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = os.getenv("CACHE_REDIS_HOST")
        app.config["CACHE_REDIS_PORT"] = os.getenv("CACHE_REDIS_PORT", 6379)

    return app


def cleanup_app(app: Flask):
    pass


def serve_app(app: Flask, cache: Cache, services: Services,
              storage: SchedulingStorage, config: Dict[str, Any],
              mount_point: str = '/scheduling',
              log_handler: StreamHandler = None):
    register_views(app, cache, services, storage)


###############################################################################
# Internals
###############################################################################
def register_views(app: Flask, cache: Cache, services,
                   storage: SchedulingStorage):
    pass


def patch_request(bp: Blueprint, services, storage: SchedulingStorage):
    """
    Configure each request by attaching the services and storage
    and detach both once when the request is finished
    """
    @bp.before_request
    def prepare_request():
        request.services = services
        request.storage = storage

        @after_this_request
        def clean_request(response: Response):
            request.services = None
            request.storage = None
            return response



# -*- coding: utf-8 -*-
from enum import Enum
from typing import Any, List, NoReturn, Union
import uuid
from uuid import UUID

from chaosplt_relational_storage.db import Base, get_secret_key
from dateparser import parse
from sqlalchemy import Column, DateTime, Integer, String
from sqlalchemy import Enum as EnumType
from sqlalchemy.orm.session import Session
from sqlalchemy_utils import EncryptedType, UUIDType
from sqlalchemy_utils import JSONType as JSONB
from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine

__all__ = ["Schedule"]


class ScheduleStatus(Enum):
    created = "created"
    pending = "pending"
    active = "active"
    completed = "completed"
    cancelled = "cancelled"
    aborted = "aborted"
    retry = "retry"
    unknown = "unknown"

    @staticmethod
    def is_valid_status(schedule_type: str) -> bool:
        return schedule_type in (
            ScheduleStatus.created.value,
            ScheduleStatus.pending.value,
            ScheduleStatus.active.value,
            ScheduleStatus.completed.value,
            ScheduleStatus.cancelled.value,
            ScheduleStatus.aborted.value,
            ScheduleStatus.retry.value,
            ScheduleStatus.unknown.value
        )

    @staticmethod
    def get_status(schedule_type: str) -> 'ScheduleStatus':
        if schedule_type == ScheduleStatus.created.value:
            return ScheduleStatus.created

        if schedule_type == ScheduleStatus.pending.value:
            return ScheduleStatus.pending

        if schedule_type == ScheduleStatus.active.value:
            return ScheduleStatus.active

        if schedule_type == ScheduleStatus.completed.value:
            return ScheduleStatus.completed

        if schedule_type == ScheduleStatus.cancelled.value:
            return ScheduleStatus.cancelled

        if schedule_type == ScheduleStatus.aborted.value:
            return ScheduleStatus.aborted

        if schedule_type == ScheduleStatus.retry.value:
            return ScheduleStatus.retry

        if schedule_type == ScheduleStatus.unknown.value:
            return ScheduleStatus.unknown


class Schedule(Base):  # type: ignore
    __tablename__ = "schedule"

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    user_id = Column(UUIDType(binary=False), nullable=False, index=True)
    org_id = Column(
        UUIDType(binary=False), nullable=False, index=True)
    workspace_id = Column(
        UUIDType(binary=False), nullable=False, index=True)
    experiment_id = Column(
        UUIDType(binary=False), nullable=False, index=True)
    token_id = Column(UUIDType(binary=False), nullable=False)
    job_id = Column(UUIDType(binary=False), nullable=True)
    scheduled = Column(DateTime(), nullable=False)
    repeat = Column(Integer, nullable=True)
    interval = Column(Integer, nullable=True)
    cron = Column(String, nullable=True)
    status = Column(EnumType(ScheduleStatus), default=ScheduleStatus.unknown)
    results = Column(EncryptedType(JSONB, get_secret_key, AesEngine, 'pkcs5'))
    settings = Column(EncryptedType(JSONB, get_secret_key, AesEngine, 'pkcs5'))
    configuration = Column(
        EncryptedType(JSONB, get_secret_key, AesEngine, 'pkcs5'))
    secrets = Column(EncryptedType(JSONB, get_secret_key, AesEngine, 'pkcs5'))

    @staticmethod
    def load(user_id: Union[UUID, str], schedule_id: Union[UUID, str],
             session: Session) -> 'Schedule':
        return session.query(Schedule).\
            filter_by(id=schedule_id).\
            filter_by(user_id=user_id).first()

    @staticmethod
    def load_by_user(user_id: Union[UUID, str],
                     session: Session) -> List['Schedule']:
        return session.query(Schedule).\
            filter_by(user_id=user_id).all()

    @staticmethod
    def create(user_id: Union[UUID, str], org_id: Union[UUID, str],
               workspace_id: Union[UUID, str], experiment_id: Union[UUID, str],
               token_id: Union[UUID, str], scheduled_at: str, repeat: int,
               interval: int, cron: str, settings: Any, configuration: Any,
               secrets: Any, session: Session) -> 'Schedule':

        scheduled = parse(scheduled_at)
        schedule = Schedule(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            status=ScheduleStatus.created,
            scheduled=scheduled,
            repeat=repeat or None,
            interval=interval or None,
            cron=cron or None,
            settings=settings,
            configuration=configuration,
            secrets=secrets
        )
        session.add(schedule)
        return schedule

    @staticmethod
    def delete(schedule_id: Union[UUID, str], session: Session) -> NoReturn:
        schedule = session.query(Schedule).filter_by(
            id=schedule_id).first()

        if schedule:
            session.delete(schedule)

    @staticmethod
    def set_job_id(schedule_id: Union[UUID, str], job_id: Union[UUID, str],
                   session: Session) -> NoReturn:
        schedule = session.query(Schedule).filter_by(
            id=schedule_id).first()
        schedule.job_id = job_id

    @staticmethod
    def set_status(schedule_id: Union[UUID, str], status: ScheduleStatus,
                   session: Session) -> NoReturn:
        schedule = session.query(Schedule).filter_by(
            id=schedule_id).first()
        schedule.status = status

    @staticmethod
    def list_by_state(status: ScheduleStatus,
                      session: Session) -> List['Schedule']:
        return session.query(Schedule).filter_by(status=status).all()



# -*- coding: utf-8 -*-
from typing import Any, Dict, Union
from uuid import UUID

from flask import Flask, request, Request
from flask_login import LoginManager
from flask_jwt_extended import JWTManager, verify_jwt_in_request, \
    current_user as api_user

__all__ = ["setup_jwt", "setup_login"]


def setup_jwt(app: Flask) -> JWTManager:
    jwt = JWTManager(app)

    @jwt.user_claims_loader
    def add_claims(identity: Union[UUID, str]) -> Dict[str, Any]:
        return {
            'user_id': identity
        }

    @jwt.user_loader_callback_loader
    def user_loader(identity: Union[UUID, str]):
        return request.services.account.registration.get(identity)

    return jwt


def setup_login(app: Flask, from_session: bool = False,
                from_jwt: bool = False) -> LoginManager:
    login_manager = LoginManager()
    login_manager.init_app(app)

    if from_session:
        @login_manager.user_loader
        def load_user_from_session(user_id: Union[UUID, str]):
            return request.services.account.registration.get(user_id)

    if from_jwt:
        @login_manager.request_loader
        def load_user_from_request(request: Request):
            verify_jwt_in_request()
            return api_user

    return login_manager



import uuid
from uuid import UUID

from chaosplt_scheduling.storage.model import Schedule, ScheduleStatus
from chaosplt_relational_storage.db import orm_session


def test_save_schedule(user_id: UUID, org_id: UUID, workspace_id: UUID,
                       experiment_id: UUID, token_id: UUID, scheduled: str):
    with orm_session() as session:
        sched = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        session.commit()
        assert uuid.UUID(hex=sched.id.hex) == sched.id
        assert sched.status == ScheduleStatus.created


def test_get_schedule(user_id: UUID, org_id: UUID, workspace_id: UUID,
                      experiment_id: UUID, token_id: UUID, scheduled: str):
    with orm_session() as session:
        sched = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        session.commit()

        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert sched == fetched_sched


def test_get_schedule_for_user(user_id: UUID, org_id: UUID, workspace_id: UUID,
                               experiment_id: UUID, token_id: UUID,
                               scheduled: str):
    with orm_session() as session:
        sched = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        session.commit()

        fetched_scheds = Schedule.load_by_user(user_id, session)
        assert len(fetched_scheds) == 1
        assert sched == fetched_scheds[0]


def test_delete_schedule(user_id: UUID, org_id: UUID, workspace_id: UUID,
                         experiment_id: UUID, token_id: UUID,
                         scheduled: str):
    with orm_session() as session:
        sched = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        session.commit()

        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert sched == fetched_sched

        Schedule.delete(sched.id, session)
        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert fetched_sched == None


def test_set_job_id(user_id: UUID, org_id: UUID, workspace_id: UUID,
                    experiment_id: UUID, token_id: UUID,
                    scheduled: str):
    with orm_session() as session:
        sched = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        session.commit()

        before_job_id = sched.job_id

        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert sched == fetched_sched

        job_id = uuid.uuid4()
        Schedule.set_job_id(sched.id, job_id, session)
        session.commit()

        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert before_job_id != fetched_sched.job_id


def test_set_status(user_id: UUID, org_id: UUID, workspace_id: UUID,
                    experiment_id: UUID, token_id: UUID,
                    scheduled: str):
    with orm_session() as session:
        sched = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        session.commit()

        before_status = sched.status

        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert sched == fetched_sched

        job_id = uuid.uuid4()
        Schedule.set_status(sched.id, ScheduleStatus.completed, session)
        session.commit()

        fetched_sched = Schedule.load(user_id, sched.id, session)
        assert before_status != fetched_sched.status
        assert fetched_sched.status == ScheduleStatus.completed


def test_list_by_state(user_id: UUID, org_id: UUID, workspace_id: UUID,
                       experiment_id: UUID, token_id: UUID,
                       scheduled: str):
    with orm_session() as session:
        sched1 = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        sched2 = Schedule.create(
            user_id=user_id,
            org_id=org_id,
            workspace_id=workspace_id,
            experiment_id=experiment_id,
            token_id=token_id,
            scheduled_at=scheduled,
            interval=None,
            cron=None,
            repeat=None,
            settings=None,
            configuration=None,
            secrets=None,
            session=session
        )
        sched2.status = ScheduleStatus.retry
        session.commit()

        scheds = Schedule.list_by_state(ScheduleStatus.created, session)
        assert len(scheds) == 1
        assert scheds[0] == sched1

        scheds = Schedule.list_by_state(ScheduleStatus.retry, session)
        assert len(scheds) == 1
        assert scheds[0] == sched2

        scheds = Schedule.list_by_state(ScheduleStatus.completed, session)
        assert len(scheds) == 0



from abc import ABC, abstractmethod
from typing import Any, List, NoReturn, Union
from uuid import UUID

from chaosplt_scheduling.model import Schedule

__all__ = ["BaseSchedulingService", "BaseSchedulingStorage"]


class BaseSchedulingService(ABC):
    @abstractmethod
    def get(self, user_id: Union[UUID, str],
            schedule_id: Union[UUID, str]) -> Schedule:
        """
        Lookup a schedule by its identifier for the given user
        """
        return self.load(user_id, schedule_id)

    @abstractmethod
    def create(self, user_id: Union[UUID, str], org_id: Union[UUID, str],
               workspace_id: Union[UUID, str], experiment_id: Union[UUID, str],
               token_id: Union[UUID, str], scheduled_at: str,
               interval: int = None, repeat: int = None, cron: str = None,
               settings: Any = None, configuration: Any = None,
               secrets: Any = None) -> Schedule:
        """
        Create a new schedule with the given context

        Return the new schedule
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, schedule_id: Union[UUID, str]) -> NoReturn:
        """
        Delete a schedule
        """
        raise NotImplementedError()

    @abstractmethod
    def get_by_user(self, user_id: Union[UUID, str]) -> List[Schedule]:
        """
        Get all schedules created by a given user
        """
        raise NotImplementedError()

    @abstractmethod
    def job_started(self, schedule_id: Union[UUID, str],
                    job_id: Union[UUID, str]) -> NoReturn:
        """
        Mark the job's status as started
        """
        raise NotImplementedError()


class BaseSchedulingStorage:
    def release(self) -> NoReturn:
        raise NotImplementedError()



from datetime import datetime
from typing import Any, Dict, List
from uuid import UUID

import attr

__all__ = ["Schedule"]


@attr.s
class Schedule:
    id: UUID = attr.ib()
    user_id: UUID = attr.ib()
    org_id: UUID = attr.ib()
    workspace_id: UUID = attr.ib()
    experiment_id: UUID = attr.ib()
    token_id: UUID = attr.ib()
    scheduled: datetime = attr.ib()
    status: str = attr.ib()
    job_id: UUID = attr.ib(default=None)
    repeat: int = attr.ib(default=None)
    interval: int = attr.ib(default=None)
    cron: str = attr.ib(default=None)
    settings: Dict[str, Any] = attr.ib(default=None)
    configuration: Dict[str, Any] = attr.ib(default=None)
    secrets: Dict[str, Any] = attr.ib(default=None)
    results: List[Dict[str, Any]] = attr.ib(default=None)



from datetime import datetime
from typing import Any

from chaosplt_grpc import remote_channel
from chaosplt_grpc.scheduler.client import schedule_experiment, \
    cancel_experiment

__all__ = ["SchedulerService"]


class SchedulerService:
    def __init__(self, config):
        grpc_config = config["grpc"]["scheduler"]
        self.scheduler_addr = grpc_config["address"]

    def release(self):
        pass

    def schedule(self, schedule_id: str, user_id: str, org_id: str,
                 workspace_id: str, experiment_id: str, token_id: str,
                 token: str, scheduled: datetime, experiment: Any,
                 interval: int, repeat: int, cron: str, settings: Any,
                 configuration: Any, secrets: Any) -> str:
        """
        Schedule a new execution with the given context

        Return the job identifier for the execution.
        """
        with remote_channel(self.scheduler_addr) as channel:
            return schedule_experiment(
                channel, schedule_id, user_id, org_id, workspace_id,
                experiment_id, token_id, token, scheduled, experiment,
                interval, repeat, cron, settings, configuration, secrets)

    def cancel(self, job_id: str):
        """
        Cancel the given job so that future executions do not take place.
        """
        with remote_channel(self.scheduler_addr) as channel:
            return cancel_experiment(channel, job_id)



from flask import Flask
from flask_caching import Cache

__all__ = ["setup_cache"]


def setup_cache(app: Flask) -> Cache:
    """
    Initialize the application's cache.
    """
    return Cache(app, config=app.config)



from typing import Any, List, NoReturn, Union
from uuid import UUID

from chaosplt_scheduling.model import Schedule
from chaosplt_relational_storage import RelationalStorage
from chaosplt_relational_storage.db import orm_session

from .interface import BaseSchedulingService
from .model import ScheduleStatus, Schedule as ScheduleModel


class SchedulingService(BaseSchedulingService):
    def __init__(self, driver: RelationalStorage):
        self.driver = driver

    def get(self, user_id: Union[UUID, str],
            schedule_id: Union[UUID, str]) -> Schedule:
        with orm_session() as session:
            schedule = ScheduleModel.load(
                user_id, schedule_id, session=session)
            if not schedule:
                return

            return Schedule(
                id=schedule.id,
                user_id=schedule.user_id,
                org_id=schedule.org_id,
                workspace_id=schedule.workspace_id,
                experiment_id=schedule.experiment_id,
                token_id=schedule.token_id,
                status=schedule.status.value,
                scheduled=schedule.scheduled,
                repeat=schedule.repeat,
                interval=schedule.interval,
                cron=schedule.cron,
                settings=schedule.settings,
                configuration=schedule.configuration,
                secrets=schedule.secrets,
                results=schedule.results
            )

    def create(self, user_id: Union[UUID, str], org_id: Union[UUID, str],
               workspace_id: Union[UUID, str], experiment_id: Union[UUID, str],
               token_id: Union[UUID, str], scheduled_at: str,
               interval: int = None, repeat: int = None, cron: str = None,
               settings: Any = None, configuration: Any = None,
               secrets: Any = None) -> Schedule:
        with orm_session() as session:
            schedule = ScheduleModel.create(
                user_id, org_id, workspace_id, experiment_id, token_id,
                scheduled_at, interval=interval, repeat=repeat, cron=cron,
                settings=settings, configuration=configuration,
                secrets=secrets, session=session)
            session.flush()

            return Schedule(
                id=schedule.id,
                user_id=schedule.user_id,
                org_id=schedule.org_id,
                workspace_id=schedule.workspace_id,
                experiment_id=schedule.experiment_id,
                token_id=schedule.token_id,
                status=schedule.status.value,
                scheduled=schedule.scheduled,
                repeat=schedule.repeat,
                interval=schedule.interval,
                cron=schedule.cron,
                settings=schedule.settings,
                configuration=schedule.configuration,
                secrets=schedule.secrets
            )

    def delete(self, schedule_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            ScheduleModel.delete(schedule_id, session=session)

    def get_by_user(self, user_id: Union[UUID, str]) -> List[Schedule]:
        with orm_session() as session:
            schedules = []

            candidates = ScheduleModel.load_by_user(
                user_id, session=session)
            for schedule in candidates:
                schedules.append(
                    Schedule(
                        id=schedule.id,
                        user_id=schedule.user_id,
                        org_id=schedule.org_id,
                        workspace_id=schedule.workspace_id,
                        experiment_id=schedule.experiment_id,
                        token_id=schedule.token_id,
                        status=schedule.status.value,
                        scheduled=schedule.scheduled,
                        repeat=schedule.repeat,
                        interval=schedule.interval,
                        cron=schedule.cron,
                        settings=schedule.settings,
                        configuration=schedule.configuration,
                        secrets=schedule.secrets,
                        results=schedule.results
                    )
                )
            return schedules

    def job_started(self, schedule_id: Union[UUID, str],
                    job_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            ScheduleModel.set_status(
                schedule_id, ScheduleStatus.pending, session=session)
            ScheduleModel.set_job_id(schedule_id, job_id, session=session)
